@@ -3,7 +3,7 @@
 #
 # SPDX-License-Identifier: Apache-2.0 OR GPL-2.0-or-later
 # Dulwich is dual-licensed under the Apache License, Version 2.0 and the GNU
-# General Public License as public by the Free Software Foundation; version 2.0
+# General Public License as published by the Free Software Foundation; version 2.0
 # or (at your option) any later version. You can redistribute it and/or
 # modify it under the terms of either of these two licenses.
 #
@@ -33,7 +33,9 @@ from urllib.parse import urlparse
 
 import dulwich
 from dulwich import client
+from dulwich.bundle import create_bundle_from_repo, write_bundle
 from dulwich.client import (
+    BundleClient,
     FetchPackResult,
     GitProtocolError,
     HangupException,
@@ -59,7 +61,7 @@ from dulwich.client import (
     parse_rsync_url,
 )
 from dulwich.config import ConfigDict
-from dulwich.objects import Commit, Tree
+from dulwich.objects import Blob, Commit, Tree
 from dulwich.pack import pack_objects_to_data, write_pack_data, write_pack_objects
 from dulwich.protocol import DEFAULT_GIT_PROTOCOL_VERSION_FETCH, TCP_GIT_PORT, Protocol
 from dulwich.repo import MemoryRepo, Repo
@@ -527,6 +529,30 @@ class TestGetTransportAndPath(TestCase):
         self.assertIsInstance(c, LocalGitClient)
         self.assertEqual("foo.bar/baz", path)
 
+    def test_ssh_with_config(self) -> None:
+        # Test that core.sshCommand from config is passed to SSHGitClient
+        from dulwich.config import ConfigDict
+
+        config = ConfigDict()
+        c, path = get_transport_and_path(
+            "ssh://git@github.com/user/repo.git", config=config
+        )
+        self.assertIsInstance(c, SSHGitClient)
+        self.assertIsNone(c.ssh_command)
+
+        config.set((b"core",), b"sshCommand", b"custom-ssh -o CustomOption=yes")
+
+        c, path = get_transport_and_path(
+            "ssh://git@github.com/user/repo.git", config=config
+        )
+        self.assertIsInstance(c, SSHGitClient)
+        self.assertEqual("custom-ssh -o CustomOption=yes", c.ssh_command)
+
+        # Test rsync-style URL also gets the config
+        c, path = get_transport_and_path("git@github.com:user/repo.git", config=config)
+        self.assertIsInstance(c, SSHGitClient)
+        self.assertEqual("custom-ssh -o CustomOption=yes", c.ssh_command)
+
     @skipIf(sys.platform != "win32", "Behaviour only happens on windows.")
     def test_local_abs_windows_path(self) -> None:
         c, path = get_transport_and_path("C:\\foo.bar\\baz")
@@ -818,6 +844,33 @@ class SSHGitClientTests(TestCase):
         test_client = SSHGitClient("git.samba.org", ssh_command="ssh -o Option1=Value1")
         self.assertEqual(test_client.ssh_command, "ssh -o Option1=Value1")
 
+    def test_ssh_command_config(self) -> None:
+        # Test core.sshCommand config setting
+        from dulwich.config import ConfigDict
+
+        # No config, no environment - should be None
+        self.overrideEnv("GIT_SSH", None)
+        self.overrideEnv("GIT_SSH_COMMAND", None)
+        test_client = SSHGitClient("git.samba.org")
+        self.assertIsNone(test_client.ssh_command)
+
+        # Config with core.sshCommand
+        config = ConfigDict()
+        config.set((b"core",), b"sshCommand", b"ssh -o StrictHostKeyChecking=no")
+        test_client = SSHGitClient("git.samba.org", config=config)
+        self.assertEqual(test_client.ssh_command, "ssh -o StrictHostKeyChecking=no")
+
+        # ssh_command parameter takes precedence over config
+        test_client = SSHGitClient(
+            "git.samba.org", config=config, ssh_command="custom-ssh"
+        )
+        self.assertEqual(test_client.ssh_command, "custom-ssh")
+
+        # Environment variables take precedence over config when no ssh_command parameter
+        self.overrideEnv("GIT_SSH_COMMAND", "/usr/bin/ssh -v")
+        test_client = SSHGitClient("git.samba.org", config=config)
+        self.assertEqual(test_client.ssh_command, "/usr/bin/ssh -v")
+
 
 class ReportStatusParserTests(TestCase):
     def test_invalid_pack(self) -> None:
@@ -948,8 +1001,10 @@ class LocalGitClientTests(TestCase):
         self.addCleanup(tear_down_repo, local)
 
         client = LocalGitClient()
-        refs = client.get_refs(local.path)
-        self.assertDictEqual(local.refs.as_dict(), refs)
+        result = client.get_refs(local.path)
+        self.assertDictEqual(local.refs.as_dict(), result.refs)
+        # Check that symrefs are detected correctly
+        self.assertIn(b"HEAD", result.symrefs)
 
     def send_and_verify(self, branch, local, target) -> None:
         """Send branch from local to remote repository and verify it worked."""
@@ -970,6 +1025,137 @@ class LocalGitClientTests(TestCase):
         self.assertEqual(obj_local, obj_target)
 
 
+class BundleClientTests(TestCase):
+    def setUp(self) -> None:
+        super().setUp()
+        self.tempdir = tempfile.mkdtemp()
+        self.addCleanup(shutil.rmtree, self.tempdir)
+
+    def _create_test_bundle(self):
+        """Create a test bundle file and return its path."""
+        # Create a simple repository
+        repo = MemoryRepo()
+
+        # Create some objects
+        blob = Blob.from_string(b"Hello world")
+        repo.object_store.add_object(blob)
+
+        tree = Tree()
+        tree.add(b"hello.txt", 0o100644, blob.id)
+        repo.object_store.add_object(tree)
+
+        commit = Commit()
+        commit.tree = tree.id
+        commit.message = b"Initial commit"
+        commit.author = commit.committer = b"Test User <test@example.com>"
+        commit.commit_time = commit.author_time = 1234567890
+        commit.commit_timezone = commit.author_timezone = 0
+        repo.object_store.add_object(commit)
+
+        repo.refs[b"refs/heads/master"] = commit.id
+
+        # Create bundle
+        bundle = create_bundle_from_repo(repo)
+
+        # Write bundle to file
+        bundle_path = os.path.join(self.tempdir, "test.bundle")
+        with open(bundle_path, "wb") as f:
+            write_bundle(f, bundle)
+
+        return bundle_path, repo
+
+    def test_is_bundle_file(self) -> None:
+        """Test bundle file detection."""
+        bundle_path, _ = self._create_test_bundle()
+
+        # Test positive case
+        self.assertTrue(BundleClient._is_bundle_file(bundle_path))
+
+        # Test negative case - regular file
+        regular_file = os.path.join(self.tempdir, "regular.txt")
+        with open(regular_file, "w") as f:
+            f.write("not a bundle")
+        self.assertFalse(BundleClient._is_bundle_file(regular_file))
+
+        # Test negative case - non-existent file
+        self.assertFalse(BundleClient._is_bundle_file("/non/existent/file"))
+
+    def test_get_refs(self) -> None:
+        """Test getting refs from bundle."""
+        bundle_path, _ = self._create_test_bundle()
+
+        client = BundleClient()
+        result = client.get_refs(bundle_path)
+
+        self.assertIn(b"refs/heads/master", result.refs)
+        self.assertEqual(result.symrefs, {})
+
+    def test_fetch_pack(self) -> None:
+        """Test fetching pack from bundle."""
+        bundle_path, source_repo = self._create_test_bundle()
+
+        client = BundleClient()
+        pack_data = BytesIO()
+
+        def determine_wants(refs):
+            return list(refs.values())
+
+        class MockGraphWalker:
+            def next(self):
+                return None
+
+            def ack(self, sha):
+                pass
+
+        result = client.fetch_pack(
+            bundle_path, determine_wants, MockGraphWalker(), pack_data.write
+        )
+
+        # Verify we got refs back
+        self.assertIn(b"refs/heads/master", result.refs)
+
+        # Verify pack data was written
+        self.assertGreater(len(pack_data.getvalue()), 0)
+
+    def test_fetch(self) -> None:
+        """Test fetching from bundle into target repo."""
+        bundle_path, source_repo = self._create_test_bundle()
+
+        client = BundleClient()
+        target_repo = MemoryRepo()
+
+        result = client.fetch(bundle_path, target_repo)
+
+        # Verify refs were imported
+        self.assertIn(b"refs/heads/master", result.refs)
+
+        # Verify objects were imported
+        master_id = result.refs[b"refs/heads/master"]
+        self.assertIn(master_id, target_repo.object_store)
+
+        # Verify the commit object is correct
+        commit = target_repo.object_store[master_id]
+        self.assertEqual(commit.message, b"Initial commit")
+
+    def test_send_pack_not_supported(self) -> None:
+        """Test that send_pack raises NotImplementedError."""
+        bundle_path, _ = self._create_test_bundle()
+
+        client = BundleClient()
+
+        with self.assertRaises(NotImplementedError):
+            client.send_pack(bundle_path, None, None)
+
+    def test_get_transport_and_path_bundle(self) -> None:
+        """Test that get_transport_and_path detects bundle files."""
+        bundle_path, _ = self._create_test_bundle()
+
+        client, path = get_transport_and_path(bundle_path)
+
+        self.assertIsInstance(client, BundleClient)
+        self.assertEqual(path, bundle_path)
+
+
 class HttpGitClientTests(TestCase):
     def test_get_url(self) -> None:
         base_url = "https://github.com/jelmer/dulwich"
@@ -1243,6 +1429,166 @@ class HttpGitClientTests(TestCase):
         with self.assertRaises(GitProtocolError, msg=error_msg):
             client.fetch_pack(b"/", check_heads, None, None)
 
+    def test_fetch_pack_dumb_http(self) -> None:
+        import zlib
+
+        from urllib3.response import HTTPResponse
+
+        # Mock responses for dumb HTTP
+        info_refs_content = (
+            b"0123456789abcdef0123456789abcdef01234567\trefs/heads/master\n"
+        )
+        head_content = b"ref: refs/heads/master"
+
+        # Create a blob object for testing
+        blob_content = b"Hello, dumb HTTP!"
+        blob_sha = b"0123456789abcdef0123456789abcdef01234567"
+        blob_hex = blob_sha.decode("ascii")
+        blob_obj_data = (
+            b"blob " + str(len(blob_content)).encode() + b"\x00" + blob_content
+        )
+        blob_compressed = zlib.compress(blob_obj_data)
+
+        responses = {
+            "/HEAD": {
+                "status": 200,
+                "content": head_content,
+                "content_type": "text/plain",
+            },
+            "/git-upload-pack": {
+                "status": 404,
+                "content": b"Not Found",
+                "content_type": "text/plain",
+            },
+            "/info/refs": {
+                "status": 200,
+                "content": info_refs_content,
+                "content_type": "text/plain",
+            },
+            f"/objects/{blob_hex[:2]}/{blob_hex[2:]}": {
+                "status": 200,
+                "content": blob_compressed,
+                "content_type": "application/octet-stream",
+            },
+        }
+
+        class PoolManagerMock:
+            def __init__(self) -> None:
+                self.headers: dict[str, str] = {}
+
+            def request(
+                self,
+                method,
+                url,
+                fields=None,
+                headers=None,
+                redirect=True,
+                preload_content=True,
+            ):
+                # Extract path from URL
+                from urllib.parse import urlparse
+
+                parsed = urlparse(url)
+                path = parsed.path.rstrip("/")
+
+                # Find matching response
+                for pattern, resp_data in responses.items():
+                    if path.endswith(pattern):
+                        return HTTPResponse(
+                            body=BytesIO(resp_data["content"]),
+                            headers={
+                                "Content-Type": resp_data.get(
+                                    "content_type", "text/plain"
+                                )
+                            },
+                            request_method=method,
+                            request_url=url,
+                            preload_content=preload_content,
+                            status=resp_data["status"],
+                        )
+
+                # Default 404
+                return HTTPResponse(
+                    body=BytesIO(b"Not Found"),
+                    headers={"Content-Type": "text/plain"},
+                    request_method=method,
+                    request_url=url,
+                    preload_content=preload_content,
+                    status=404,
+                )
+
+        def determine_wants(heads, **kwargs):
+            # heads contains the refs with SHA values, just return the SHA we want
+            return [heads[b"refs/heads/master"]]
+
+        received_data = []
+
+        def pack_data_handler(data):
+            # Collect pack data
+            received_data.append(data)
+
+        clone_url = "https://git.example.org/repo.git/"
+        client = HttpGitClient(clone_url, pool_manager=PoolManagerMock(), config=None)
+
+        # Mock graph walker that says we don't have anything
+        class MockGraphWalker:
+            def ack(self, sha):
+                return []
+
+        graph_walker = MockGraphWalker()
+
+        result = client.fetch_pack(
+            b"/", determine_wants, graph_walker, pack_data_handler
+        )
+
+        # Verify we got the refs
+        expected_sha = blob_hex.encode("ascii")
+        self.assertEqual({b"refs/heads/master": expected_sha}, result.refs)
+
+        # Verify we received pack data
+        self.assertTrue(len(received_data) > 0)
+        pack_data = b"".join(received_data)
+        self.assertTrue(len(pack_data) > 0)
+
+        # The pack should be valid pack format
+        self.assertTrue(pack_data.startswith(b"PACK"))
+        # Pack header: PACK + version (4 bytes) + num objects (4 bytes)
+        self.assertEqual(pack_data[4:8], b"\x00\x00\x00\x02")  # version 2
+        self.assertEqual(pack_data[8:12], b"\x00\x00\x00\x01")  # 1 object
+
+    def test_timeout_configuration(self) -> None:
+        """Test that timeout parameter is properly configured."""
+        url = "https://github.com/jelmer/dulwich"
+        timeout = 30
+
+        c = HttpGitClient(url, timeout=timeout)
+        self.assertEqual(c._timeout, timeout)
+
+    def test_timeout_from_config(self) -> None:
+        """Test that timeout can be configured via git config."""
+        from dulwich.config import ConfigDict
+
+        url = "https://github.com/jelmer/dulwich"
+        config = ConfigDict()
+        config.set((b"http",), b"timeout", b"25")
+
+        c = HttpGitClient(url, config=config)
+        # The timeout should be set on the pool manager
+        # Since we can't easily access the timeout from the pool manager,
+        # we just verify the client was created successfully
+        self.assertIsNotNone(c.pool_manager)
+
+    def test_timeout_parameter_precedence(self) -> None:
+        """Test that explicit timeout parameter takes precedence over config."""
+        from dulwich.config import ConfigDict
+
+        url = "https://github.com/jelmer/dulwich"
+        config = ConfigDict()
+        config.set((b"http",), b"timeout", b"25")
+
+        c = HttpGitClient(url, config=config, timeout=15)
+        self.assertEqual(c._timeout, 15)
+
 
 class TCPGitClientTests(TestCase):
     def test_get_url(self) -> None:
@@ -1502,6 +1848,32 @@ class DefaultUrllib3ManagerTest(TestCase):
         manager = default_urllib3_manager(config=None, cert_reqs="CERT_NONE")
         self.assertEqual(manager.connection_pool_kw["cert_reqs"], "CERT_NONE")
 
+    def test_timeout_parameter(self) -> None:
+        """Test that timeout parameter is passed to urllib3 manager."""
+        timeout = 30
+        manager = default_urllib3_manager(config=None, timeout=timeout)
+        self.assertEqual(manager.connection_pool_kw["timeout"], timeout)
+
+    def test_timeout_from_config(self) -> None:
+        """Test that timeout can be configured via git config."""
+        from dulwich.config import ConfigDict
+
+        config = ConfigDict()
+        config.set((b"http",), b"timeout", b"25")
+
+        manager = default_urllib3_manager(config=config)
+        self.assertEqual(manager.connection_pool_kw["timeout"], 25)
+
+    def test_timeout_parameter_precedence(self) -> None:
+        """Test that explicit timeout parameter takes precedence over config."""
+        from dulwich.config import ConfigDict
+
+        config = ConfigDict()
+        config.set((b"http",), b"timeout", b"25")
+
+        manager = default_urllib3_manager(config=config, timeout=15)
+        self.assertEqual(manager.connection_pool_kw["timeout"], 15)
+
 
 class SubprocessSSHVendorTests(TestCase):
     def setUp(self) -> None:
